--
-- Author: Markus Kusano
--
-- Happens-before error fixer.
--
-- The input is a C file and a list of happens before edges which must be
-- satisfied.
--
-- ./HBFixer <input C file> <input edge file>
--
-- The fixed C file is output to stdout
--
-- The edge file is defined loosly as a conjunctive normal form of happens
-- before edges which must be satisfied.
--
-- Each line is a disjunction of edges to be satisfied. All of the lines
-- together are a conjunction of clauses to be satisfied.
--
-- Satisfied is used loosly here: when a disjunction is more than a unit clause
-- (i.e., an disjunction of multiple happens before edges) it is assumed that
-- either one can be selected without causing a conflict in any other clause.
--
-- See EdgeListParser.hs for a description of the format of the edge file.
--

import System.Environment (getArgs)
import System.Exit (exitFailure)
import Control.Monad (when)
import Data.Maybe (isNothing, fromJust)
import Data.Char (isSpace)
import Data.List
import qualified Data.Set as Set
import EdgeListParser
import Control.Monad

-- A CNF formula of edges. Each sublist is a clause (OR'd) and each of the
-- clauses are AND'd together.
type EdgeDisjunc = [Edge]
type EdgeCNF = [EdgeDisjunc]

-- String describing usage information
usageStr :: String
usageStr = "Usage: HBFixer <input C file> <input event file>"

-- Extract the first and second item from the passed list into a pair. If there
-- are not exactly 2 items in the list then the program returns Nothing
processArgs :: [String] -> Maybe (String, String)
processArgs (x:y:_) = Just (x, y)
processArgs _ = Nothing

-- Dumps the passed string to stderr. Displays usage information. Exits with -1
-- code
exitError :: String -> IO ()
exitError s = do 
  putStrLn $ "[ERROR] " ++ s
  putStrLn $ usageStr
  exitFailure

-- The input C source code is first labeled with line numbers. This allows for
-- new lines to be inserted (with dummy line numbers) which will not affect
-- subsequent modifications.
--
-- For example, consider that we want to perform a modification on line 12 and
-- line 15 in the original file. If we do not save the initial line numbers,
-- then after the modifications on line 12, line 15 will not longer be on the
-- 15th line.
--
-- Negative numbers used for the line number as assumed to be new insertions to
-- the original file
type SourceLine = (Int, String)
type SourceFile = [SourceLine]

-- returns true if the string on the source line matches "#include <pthread.h>"
isPthreadH :: SourceLine -> Bool
-- First remove all white space and then check if the include matches
isPthreadH (_,s) = "#include<pthread.h>" == filter (not . isSpace) s 

-- Return the name of a condition variable given the passed Int
getCondString :: Int -> String
getCondString i = "cond" ++ (show i)

-- Return the name of a mutex given the passed Int
getMutexString :: Int -> String
getMutexString i = "mutex" ++ (show i)

-- Creats a string for the boolean protecting a condition variable
getCondBoolString :: Int -> String
getCondBoolString i = getCondString i ++ "_bool"

-- Given a disjunction of happens-before, modify the source code to satisfy the
-- edges. The passed Int should be a unique identifier for the clause. It is
-- used to give names to mutexes/condition variables
satisfyClause :: Int -> EdgeDisjunc -> SourceFile -> SourceFile
-- A unit clause can only be satisfied by inserting a wait--signal pair
--satisfyClause i [((l1, _), (l2, _))] s = insAndInitCondWait i l1 l2 s
satisfyClause i (e:[]) s = insAndInitCondWait i (fst e) (snd e) s
-- For the 2 clause disjunction, check if we can use a mutex instead of a
-- single condition variable (this should allow for more concurrency).
-- Otherwise, simply select the first item in the clause and satisfy it using a
-- condition variable.
satisfyClause i (e1:e2:[]) s = 
  if (isMutexSat s e1 e2)
    then insAndInitCS i e1 e2 s
    else insAndInitCondWait i (fst e1) (snd e1) s
-- For the 3 or more clause disjunction, just select the first
--satisfyClause i (e:_) s = insAndInitCondWait i (fst e) (snd e) s
--satisfyClause _ [] s = s
-- For a 3 or more clause disjunction, search inside of it to see if any of the
-- sub-pairs form the mutex pattern. Otherwise, just select a term and satisfy
-- it with a cond--signal
satisfyClause i edges s = 
  case filter (isMutexSatPair s) allPairs of
  [] -> insAndInitCondWait i (fst . head $ edges) (snd . head $ edges) s
  ((e1,e2):es)  -> insAndInitCS i e1 e2 s
  --((e1,e2):es)  -> error $ "Error: CS edges:\n" ++ (show e1) ++ "\n" ++ (show e2)
  where allPairs = mkAllUniqPairs edges
        isMutexSatPair s (e,e') = isMutexSat s e e'
        _                   = False


-- Given a list, create all pairs of unique (i.e., non-replicated (e.g.,
-- (1,1))) elements
mkAllUniqPairs :: Ord a => [a] -> [(a,a)]
mkAllUniqPairs [] = []
mkAllUniqPairs es = pairs
  where allPairs = replicateM 2 es
        uniqs = filter (not . isRepPair) allPairs
        pairs = map (listToPair) uniqs

-- Return true if the list is size 2 and contains the same item twice
isRepPair :: Eq a => [a] -> Bool
isRepPair [] = False
isRepPair (e1:e2:[]) = e1 == e2
isRepPair _ = False

-- Converts a size two list to a pair, otherwise throws an error
listToPair :: [a] -> (a,a)
listToPair (a1:a2:as) = (a1, a2)
listToPair l = error $ "Error: listToPair: size " ++ (show . length $ l) ++ " list"



-- Inserts and initializes the mutex used in a critical section. 
--
-- The integers associated with each thread (i.e., the line number) are assumed
-- to be in program order. That is, if "t1" has lines 8 and 12 involved with
-- this edge then 8 is assumed to happen before 12. That is, the lock call
-- should be inserted before 8 and the unlock after 12.
--
-- The passed Int is an identifier used to name the mutex.
--
-- The mutex is statically initialized as a global.
insAndInitCS :: Int -> Edge -> Edge -> SourceFile -> SourceFile
insAndInitCS i e1 e2 s = insMutexInit mutStr
                         -- The snd element of an edge is the sink. The first
                         -- element of a statement is the source line

                         . insLockCall firstLockPos mutStr 
                         . insLockCall secondLockPos mutStr 
                         -- The first element of an edge is the source. The
                         -- first element of a statement is the source line.
                         -- Insert the unlock call after the statement
                         . insUnlockCall (firstUnlockPos + 1) mutStr 
                         . insUnlockCall (secondUnlockPos + 1) mutStr 
                         $ s
    where mutStr = getMutexString i
          -- isMutexSat ensures that the first statement of the first edge and
          -- the second statement of the second edge are from the same thread
          -- (and vice versa)
          --
          -- The line number is the second of a statement
          firstThreadLines = map getLineNum  [getSource e1, getSink e2]
          secondThreadLines = map getLineNum [getSink $ e1, getSource e2]
          -- The lock position is the minimum of the line numbers
          firstLockPos = minimum firstThreadLines
          secondLockPos = minimum secondThreadLines
          -- Unlock is the maximum line number
          firstUnlockPos = maximum firstThreadLines
          secondUnlockPos = maximum secondThreadLines



-- Returns true if the two passed edges match the mutex pattern.
--
-- This is true if for the two edges e1 and e2 the thread for the first
-- statement in e1 matches the thread for the second statement in e2. And, if
-- the thread for the second statement in e1 matches the thread for the first
-- statement in e2.
--
-- Additionally, make sure that the sink of e1 is not the source of e2 (i.e.,
-- this is not actually just a three node graph instead of 4)
--
-- The edges, along with the intra--thread edges (determined by line
-- number) must form a cycle. This ensures that the two edges represent a
-- critical section: that is, either thread one can go first or thread two can
-- go first
--
-- If the edges are between mutex lock or unlock calls then they cannot be
-- solved by the mutex pattern. This is because en-forcing, for example, an
-- unlock() before another unlock() is actually inserting a deadlock into the
-- program
--
-- In this case, we assume there is a program order causality between the
-- statements within each thread so we can insert a mutex instead of a (more
-- restrictive) condition wait--signal
isMutexSat :: SourceFile -> Edge -> Edge -> Bool
isMutexSat s e1 e2 =  (not (isDoubleLockOrUnlock s e1)) 
                   && (not (isDoubleLockOrUnlock s e2))
                   && (not (willDeadlock s e1 e2))
                   && (e1SrcThread == e2SnkThread)
                   && (e1SnkThread == e2SrcThread)
                   -- && (fourOrTwoNodes e1 e2)
                   && (fourOrTwoNodes e1 e2)
                   -- Check for a cycle
                   && ((e1SrcLine == t1MaxLine)
                      && (e1SnkLine == t2MinLine))
                   && ((e2SrcLine == t2MaxLine)
                      && (e2SnkLine == t1MinLine))
                      where e1SrcThread = (snd (fst e1))
                            e1SnkThread = (snd (snd e1))
                            e2SrcThread = (snd (fst e2))
                            e2SnkThread = (snd (snd e2))
                            e1SrcLine = (getLineNum (fst e1))
                            e1SnkLine = (getLineNum (snd e1))
                            e2SrcLine = (getLineNum (fst e2))
                            e2SnkLine = (getLineNum (snd e2))
                            -- These are split into thread1 and thread2
                            -- although they are not required to be the
                            -- linenumbers associated with each thread.
                            -- However, the first checks above ensure that when
                            -- these values are used they correspond to
                            -- matching threads
                            t1Line1 = getLineNum $ fst e1
                            t1Line2 = getLineNum $ snd e2
                            t2Line1 = getLineNum $ snd e1
                            t2Line2 = getLineNum $ fst e2
                            t1MinLine = minimum [t1Line1, t1Line2]
                            t1MaxLine = maximum [t1Line1, t1Line2]
                            t2MinLine = minimum [t2Line1, t2Line2]
                            t2MaxLine = maximum [t2Line1, t2Line2]

-- Returns true if two edges contain 4 or two unique nodes (i.e., 4/2 unique line
-- numbers)
fourOrTwoNodes :: Edge -> Edge -> Bool
fourOrTwoNodes e1 e2 = (size == 2) || (size == 4)
  where e1s = getLineNums e1
        e2s = getLineNums e2
        lns = e1s ++ e2s
        size = (Set.size . Set.fromList $ lns)

-- Returns true if two edges contain 4 or two unique nodes (i.e., 4 unique line
-- numbers)
fourNodes :: Edge -> Edge -> Bool
fourNodes e1 e2 = (size == 4)
  where e1s = getLineNums e1
        e2s = getLineNums e2
        lns = e1s ++ e2s
        size = (Set.size . Set.fromList $ lns)

getLineNums :: Edge -> [Int]
getLineNums e = (fst (fst e)) : (fst (snd e)) : []



-- Creates a new global condition variable using the passed Int (first param)
-- as a unique identifier. Creates a new mutex for the condition variable
-- similarly. Inserts cond_signal before the first integer and cond_wait before
-- the second integer. Also inserts the condition checking logic (and booleans)
-- for the condition variable.
insAndInitCondWait :: Int -> Statement -> Statement -> SourceFile -> SourceFile
insAndInitCondWait tag signalStmnt waitStmnt s = 
  insCondInit condStr 
  . insBool boolStr
  . insMutexInit mutexStr 
  -- Insert the condition wait
  . insCondWait waitLine condStr boolStr mutexStr
  . insCondSignal signalLine condStr boolStr mutexStr
  -- Insert the mutex unlock call after the signalLine. This will wrap the line
  -- in the lock-unlock section
  . insUnlockCall (signalLine + 1) mutexStr 
  $ s
  where waitLine = fst waitStmnt
        signalLine = fst signalStmnt
        mutexStr = getMutexString tag
        condStr = getCondString tag
        boolStr = getCondBoolString tag

-- Inserts a global boolean with the passed name. Statically initialized to false (0).
insBool :: String -> SourceFile -> SourceFile
insBool name s = insAfterTrue isPthreadH s (typeStr ++ " " ++ name ++ " = " ++ initStr)
  where typeStr = "int"
        initStr = "0;"

-- Inserts cond_wait logic at the passed source line. The passed strings are
-- the name of the global condition variable, boolean, and mutex associated
-- with the condition variable.
insCondWait :: Int -> String -> String -> String -> SourceFile -> SourceFile
insCondWait cwl cond bool mutex s = 
  insBeforeTrue (isSpecifiedLine cwl) s condWaitStr
    where
      condWaitStr = "pthread_mutex_lock(&" ++ mutex ++ ");\nwhile (!" ++ bool 
                    ++ ") {\n\tpthread_cond_wait(&" ++ cond ++ ", &" ++ mutex 
                    ++ ");\n}\n" ++ "pthread_mutex_unlock(&" ++ mutex ++ ");\n"

-- Inserts a lock call before the passed source line number. The passed string
-- is the name of the mutex. It is assumed to be a global (i.e., not a
-- pointer).
insLockCall :: Int -> String -> SourceFile -> SourceFile
insLockCall l mut s = 
  insBeforeTrue (isSpecifiedLine l) s lockStr
    where
      lockStr = "pthread_mutex_lock(&" ++ mut ++ ");\n"

-- Inserts a unlock call before the passed source line number. The passed string
-- is the name of the mutex. It is assumed to be a global (i.e., not a
-- pointer).
insUnlockCall :: Int -> String -> SourceFile -> SourceFile
insUnlockCall l mut s = 
  insBeforeTrue (isSpecifiedLine l) s unlockStr
    where
      unlockStr = "pthread_mutex_unlock(&" ++ mut ++ ");\n"


-- Inserts cond_signal logica at the passed source line. The passed strings
-- are the name of the global condition variable, boolean, and mutex associated
-- with the condition variable. The mutex unlock call is not added and should
-- be added manually
insCondSignal :: Int -> String -> String -> String -> SourceFile -> SourceFile
insCondSignal cwl cond bool mutex s = 
  insBeforeTrue (isSpecifiedLine cwl) s condSignalStr
    where
      condSignalStr = "pthread_mutex_lock(&" ++ mutex ++ ");\n" ++ bool 
                    ++ " = 1;\n" ++ "pthread_cond_signal(&" ++ cond 
                    ++ ");\n" 
      
-- Inserts pthread_cond_t with default static initializer after the pthread
-- #include. The passed string should be the name of the condition variable
insCondInit :: String -> SourceFile -> SourceFile
insCondInit name s = insAfterTrue isPthreadH s (typeStr ++ " " ++ name ++ " = " ++ initStr)
  where typeStr = "pthread_cond_t"
        initStr = "PTHREAD_COND_INITIALIZER;"

-- Insert pthread_mutex_t with the passed name and default static initializer.
-- Inserts after #include pthread_t
insMutexInit :: String -> SourceFile -> SourceFile
insMutexInit name s = insAfterTrue isPthreadH s (typeStr ++ " " ++ name ++ " = " ++ initStr)
  where typeStr = "pthread_mutex_t"
        initStr = "PTHREAD_MUTEX_INITIALIZER;"

-- Returns true if the passed SourceLine has the passed Int as line number
isSpecifiedLine :: Int -> SourceLine -> Bool
isSpecifiedLine i (l, _) = i == l

-- Insert the passed string after the passed predicate becomes true. The new
-- source line will have -1 as line number.
insAfterTrue :: (SourceLine -> Bool) -> SourceFile -> String -> SourceFile
insAfterTrue p f s = insAfterTrueHelper p f s []

-- Accumlate the source file before the point of insertion
insAfterTrueHelper :: (SourceLine -> Bool) -> SourceFile -> String -> SourceFile -> SourceFile
insAfterTrueHelper p ((n,l):ls) s acc = 
  case p (n,l) of
    True -> acc ++ [(n,l)] ++ [(-1, s)] ++ ls
    False -> insAfterTrueHelper p ls s (acc ++ [(n,l)])
-- In this case, we've reached the end of the source file before the predicate
-- became true
insAfterTrueHelper _ [] _ _ = error $ "insAfterTrueHelper: Line not found"

-- Insert the passed string on the line before the line where the passed
-- predicate becomes true. The new source line will have -1 as line number
insBeforeTrue :: (SourceLine -> Bool) -> SourceFile -> String -> SourceFile
insBeforeTrue p f s = insBeforeTrueHelper p f s []

-- Accumlate the source file before the point of insertion
insBeforeTrueHelper :: (SourceLine -> Bool) -> SourceFile -> String -> SourceFile -> SourceFile
insBeforeTrueHelper p ((n,l):ls) s acc = 
  case p (n,l) of
    True -> acc ++ [(-1, s)] ++ [(n,l)] ++ ls
    False -> insBeforeTrueHelper p ls s (acc ++ [(n,l)])
-- In this case, we've reached the end of the source file before the predicate
-- became true
insBeforeTrueHelper _ [] _ _ = error $ "insBeforeTrueHelper: Line not found"

-- Satisfy all the clauses in the edge CNF and return the modified source code
performMods :: EdgeCNF -> SourceFile -> SourceFile
performMods e s = performModsHelper 0 e s

-- Count which clause we are operating on to give it a unique identifier
performModsHelper :: Int -> EdgeCNF -> SourceFile -> SourceFile 
-- Satisfy each clause in the CNF formula
performModsHelper i (e:es) s = performModsHelper (i + 1) es (satisfyClause i e s)
performModsHelper _ [] s = s

-- Given a source code line number, return the corresponding source line in the
-- passed file. Crashes if the error does not exist
getSourceLine :: Int -> SourceFile -> SourceLine
getSourceLine i f = 
  case filter (isSpecifiedLine i) f of
    [] -> error "Line not found in file"
    (x:[]) -> x
    (x:xs) -> error "Line occurs multiple times in file"

-- Returns true if the passed statement is a pthread_lock() in the passed file
isPthreadLock :: SourceFile -> Statement -> Bool
isPthreadLock f s = stmntMatchesStr f s (isInfixOf "pthread_mutex_lock")
-- isPthreadLock f s = isInfixOf "pthread_mutex_lock" srcLn
--   where lineNum = getLineNum s
--         (_, srcLn) = getSourceLine lineNum f 

-- Returns true if the passed statement is a pthread_lock() in the passed file
isPthreadUnlock :: SourceFile -> Statement -> Bool
isPthreadUnlock f s = stmntMatchesStr f s (isInfixOf "pthread_mutex_unlock")
  --isInfixOf "pthread_mutex_unlock" srcLn
  --where lineNum = getLineNum s
        --(_, srcLn) = getSourceLine lineNum f 


-- Return true if the passed statement is a
-- __{CPROVER,VERIFIER}_atomic_{begin,end}
isAtomicBeginEnd :: SourceFile -> Statement -> Bool
isAtomicBeginEnd f s = 
  if isCproverAtomic || isVerifierAtomic 
    then True
    else False
    where isCproverAtomic = stmntMatchesStr f s (isInfixOf "__CPROVER_atomic")
          isVerifierAtomic = stmntMatchesStr f s (isInfixOf "__VERIFIER_atomic")

-- Extracts the passed Statement from the passed source file. Runs the passed
-- function (String -> Bool) on the statement and returns the result
stmntMatchesStr :: SourceFile -> Statement -> (String -> Bool) -> Bool
stmntMatchesStr file st func = func srcLn
  where lineNum = getLineNum st
        -- Extract the line of code
        (_, srcLn) = getSourceLine lineNum file

-- Returns true if the passed edge is a lock--unlock pair. That is, one is a
-- lock() call and the other is an unlock() call (they may be to different
-- mutexes).
-- Also returns true if either of the statements are a
-- __CPROVER_atomic_{begin,end} statement
isLockUnlockEdge :: SourceFile -> Edge -> Bool
isLockUnlockEdge f (s1, s2) =
  -- Lock --> Unlock
  if (isPthreadLock f s1) && (isPthreadUnlock f s2) 
    then True
  -- Unlock --> Lock
    else if (isPthreadUnlock f s1) && (isPthreadLock f s2) 
      then True
      else if (isAtomicBeginEnd f s1) || (isAtomicBeginEnd f s2)
        then True
        else False

-- Returns true if the edges is a lock--lock or unlock--unlock edge
isDoubleLockOrUnlock :: SourceFile -> Edge -> Bool
isDoubleLockOrUnlock f (s1, s2) =
  if (isPthreadLock f s1) && (isPthreadLock f s2)
    then True
    else if (isPthreadUnlock f s1) && (isPthreadUnlock f s2)
      then True
      else False

-- Returns true if the addition of the passed edge to the passed SourceFile
-- will cause a deadlock. In general, this is a cause of much inaccuraccy.
--
-- The current implementation considers an edge to deadlock if the
-- minimum--maximum path of the two threads involved in the two passed edges
-- does not contain any unlock calls. This is an over approximation since it
-- ignores control flow (i.e., it is flow insensitive, ignoring branch
-- statements).
--
-- This function assumes that the two passed edges are from two different
-- threads and form a mutex pattern (see isMutexSat)
willDeadlock :: SourceFile -> Edge -> Edge -> Bool
willDeadlock s e1 e2 = if isLockUnlockEdge s e1
                          -- e1 ends at the start of thread two (see below)
                          then if containsUnlock s t2MinLine t2MaxLine
                                 then True
                                 else if isLockUnlockEdge s e2
                                        -- e2 ends at the start of thread one (see below)
                                        then if containsUnlock s t1MinLine t1MaxLine
                                             then True
                                             else False
                                        else False
                          else False
  where e1SrcThread = (snd (fst e1))
        e1SnkThread = (snd (snd e1))
        e2SrcThread = (snd (fst e2))
        e2SnkThread = (snd (snd e2))
        e1SrcLine = (getLineNum (fst e1))
        e1SnkLine = (getLineNum (snd e1))
        e2SrcLine = (getLineNum (fst e2))
        e2SnkLine = (getLineNum (snd e2))
        t1Line1 = getLineNum $ fst e1
        t1Line2 = getLineNum $ snd e2
        t2Line1 = getLineNum $ snd e1
        t2Line2 = getLineNum $ fst e2
        t1MinLine = minimum [t1Line1, t1Line2]
        t1MaxLine = maximum [t1Line1, t1Line2]
        t2MinLine = minimum [t2Line1, t2Line2]
        t2MaxLine = maximum [t2Line1, t2Line2]

-- Returns true if there exists an unlock call within the passed line numbers
containsUnlock :: SourceFile -> Int -> Int -> Bool
containsUnlock s i1 i2 
    -- Range does not make sense
    | i1 > i2 = error "Error: containsUnlock: i2 < i1"
    | i1 < i2 = or $ map (isInfixOf "pthread_mutex_unlock") lns
            -- Get all the lines except for the last line in the range (we are
            -- checking if there exists an unlock anywhere _within_ the passed
            -- range 
    | i1 == i2 = False
    | otherwise = error $ "inexauhstive (containsUnlock)?" ++ (show i1) ++ " " ++ (show i2) 
      where lns = getLines s i1 (i2 - 1) 


-- Given the passed SourceFile, collect all the lines within the passed range
-- of integers
getLines :: SourceFile -> Int -> Int -> [String]
getLines s i1 i2 
    | i1 > i2 = error "Error: getLines: i1 > i2"
    | i1 < i2 = map snd (filter matches s)
            -- Returns true if the passed SourceLine is within the range
    | otherwise = error $ "inexauhstive (getLines)?" ++ (show i1) ++ " " ++ (show i2) 
      where matches :: SourceLine -> Bool
            matches (i, _) = (i >= i1) || (i <= i2)



-- Given an EdgeCNF, in each caluse in the CNF formula, move those edges which
-- are A->B where A is a lock() and B is an unlock() (or vice-versa) to the end
-- of the formula. The rest of the program choses the first items in the list.
-- As a result, this will cause non lock--unlock pairs to be chosen first.
perferNonLockUnlock :: SourceFile -> EdgeCNF -> EdgeCNF
perferNonLockUnlock f cnf = map (perferNonLockUnlock' f) cnf

-- Helper function: Moves lock--unlock edges to the end of a caluse in a CNF
-- formula
perferNonLockUnlock' :: SourceFile -> EdgeDisjunc -> EdgeDisjunc
perferNonLockUnlock' f es = others ++ lu
  where (lu, others) = partition (isLockUnlockEdge f) es

-- Given a list of possible solutions, return a CNF form of the ``best''
-- solution. A CNF is used for the case where there are multiple single
-- happens-before edge solutions. This allows the fixer to use a mutex to
-- enforce an either-or edge (i.e., A->B or B->A). 
--
-- The best solution is determined by the number of happens-before edges
-- required in the solution. Lower numbers are perfered. For the case where
-- there are more than 2 edges required in the solution, only one solution is
-- returned (i.e., the result is a conjunction of unit clauses)
getBestSolution :: [EdgeConjunct] -> EdgeCNF
getBestSolution [] = []
getBestSolution sols = 
  case (length . head $ sortedSols) of
    -- Gather all the length one solutions and return them in a single list.
    -- This means they are all in one clause
    1 -> [sizeOnes]
      where sizeOnes = concat . (takeWhile isSizeOne) $ sortedSols

    -- Otherwise, just return the first solution. We need to convert it to a
    -- CNF: it is a list of unit clauses
    _ -> map (: []) (head sortedSols)
  where sortedSols = sortBy sortListLength sols
        isSizeOne = (==) 1 . length
        sortListLength :: [a] -> [a] -> Ordering
        sortListLength a b = if (length a) < (length b)
                               then LT
                               else if (length a) > (length b)
                                 then GT
                                 else EQ

main :: IO ()
main = do 
  args <- getArgs
  -- the C file to operate on is the first argument, the file containing the
  -- events to fix is the second argument. processArgs will crash if the
  -- arguments are malformed
  let files = processArgs args
  when (isNothing files) (exitError ("Invalid number of arguments: " ++ (show (length args))))
  let (cFile, formulaFile) = fromJust files

  -- Read in the input C file and add line numbers to the lines
  sourceLines <- readFile cFile
  let sourceCode = zip ([1..] :: [Int]) $ lines sourceLines

  -- Parse the edge formula file
  formulaLines <- readFile formulaFile
  --let edges = (perferNonLockUnlock sourceCode) . parseEdgeFile $ lines formulaLines
  let edges = parseEdgeFile $ lines formulaLines

  -- Given all the potential solutions, find the ``best'' one
  let soln = (perferNonLockUnlock sourceCode) (getBestSolution edges)

  -- Given the happens-before edges which need to be satisfied, figure out the
  -- modifications required to be made on the source code
  --let mods = getModifications edges
  --let modCode = performMods edges sourceCode
  let modCode = performMods soln sourceCode

  -- Write the modified source code to file. The line numbers can be discarded
  putStr . unlines . map snd $ modCode
