-- 
-- Author: Markus Kusano
--
-- Parser for EdgeList
--
-- The edge list is defined as follows:
-- Each line in the list is a conjunctions of happens before edges which must
-- be satisifed. Each edge in the conjunction is separated by a '$':
--
-- <edge 1> $ <edge 2> $ ... $ <edge n>
--
-- Each line is a seperate conjunction. Together, they represented a
-- disjunction of formulas which must be satisfied (i.e., at least one line
-- must be satisfied
--
-- The edges are defined as follows:
--
-- <stmnt 1> , <stmnt 2>
--
-- That is, an edge is two statements separated by a comma.
--
-- Each statement is a pair: 
--
-- <line number> <thread identifier>
--
-- The line number is the line number in the C file and the thread identifier is
-- some string to uniquely identify the thread. This is used for an
-- optimization allowing for mutexes to be used instead of condition variables.
--
-- An edge '<stmnt 1>,<stmnt 2>' indicates that stmnt must be executed before
-- stmnt 2.
module EdgeListParser (ThreadId, Statement, Edge, EdgeConjunct, parseEdgeFile, getLineNum, getThreadId, getSource, getSink) where
import Text.ParserCombinators.Parsec
import Control.Applicative (some, (<$>))
import Data.Either (rights)

-- Threads are identified by some string
type ThreadId = String
-- A statement is a line number and a thread identifier
type Statement = (Int, ThreadId)
-- An edge is two statements s1 and s2 where s1 -> s2
type Edge = (Statement, Statement)
-- A conjunction of edges is a list of edges
type EdgeConjunct = [Edge]
-- A formula of  is a disjunction of conjunctions
--type EdgeForm = [EdgeConjunct]

getLineNum :: Statement -> Int
getLineNum = fst

getThreadId :: Statement -> ThreadId
getThreadId = snd

getSource :: Edge -> Statement
getSource = fst

getSink :: Edge -> Statement
getSink = snd

-- Takes a list of each line in a file and parses it into a list of
-- conjunctions of edges
parseEdgeFile :: [String] -> [EdgeConjunct]
parseEdgeFile s = rights $ map (parse line "(error)") s

-- A line is many edges separated by dollar signs (conjunction of edges, each
-- dollar sign is an AND)
--line = sepBy edge (char '$')
line :: GenParser Char st EdgeConjunct
line = do
          spaces
          e1 <- edge
          others <- remainingEdges
          return (e1 : others)

remainingEdges :: GenParser Char st [Edge]
remainingEdges = spaces >> ((char '$' >> line) <|> (return []))


-- An edge is two events separated by a comma
edge :: GenParser Char st Edge
edge = do 
          spaces
          e1 <- stmnt
          spaces 
          _ <- char ','
          spaces
          e2 <- stmnt
          return (e1,e2)


-- An event is a line number and thread ID separated by white space
stmnt :: GenParser Char st Statement
stmnt = do 
           spaces
           l <- integer
           spaces
           t <- some (noneOf ", \t\n")
           return (l, t)

-- Parse an integer
integer :: GenParser Char st Int
integer = rd <$> many1 digit
  where rd = read :: String -> Int


