-- Author: Markus Kusano
--
-- EdgeMaker parses the output of CBMC and generates a list of happens-before
-- edges which must be enforced in the program in order to make the program
-- bug-free.
--
-- This uses the set of UNSAT cores generated by the partial MAX-SAT solver.
-- Given the control-flow edges which are causeing the violation, this file
-- does the following:
--
-- 1. Generate a graph of the control-flow edges (for each interleaving)
-- For each generated graph:
-- 2. Add and intra-thread edges to the graph
-- 3. Find all the possible edges which can be added to the graph 
-- Across all graphs
-- 4. Find the happens before edges which can be added to the graph to block
-- all the interleavings (i.e., the happens-before edges which can be added
-- with conflict to block every interleaving)

module Main where 
import qualified Data.Graph as Graph
import qualified Data.Map.Strict as Map
import Control.Monad (when)
import Data.Either (rights, lefts)
import Data.Tuple (swap)
import Data.List (intersperse, find, sortBy, sort)
import Control.Monad (replicateM)
import qualified Data.Set as Set

import FormulaParser (parseFormulaFile, VarId, Formula)
import VariableParser (parseVariableFile, TId(TId), LineNum(LineNum), VertInfo, VertInfoEdge, flipEdge)
import GraphEdgeSat (GraphID, IdToNode, NodeToId, graphEdgesToBoolean, 
                     edgeToVertInfoEdge, solveFormulaUnsafe)

import Data.Maybe (maybeToList, catMaybes)

-- A map to lookup a VertInfoEdge given its Id
type IdToInfo = Map.Map VarId VertInfoEdge
--type InfoToId  = Map.Map VertInfoEdge VarId


-- Name of the formula file
formulaFile :: String
formulaFile = "formulas"

-- Name of the variables file
variableFile :: String
variableFile = "var_defs"

getVertInfoLineNum :: VertInfo -> LineNum
--getVertInfoLineNum = snd
getVertInfoLineNum (_, l, _) = l

getVertInfoTId :: VertInfo -> TId
--getVertInfoTId = fst
getVertInfoTId (t, _, _) = t

getVertInfoRW :: VertInfo -> Bool
getVertInfoRW (_, _, w) = w


-- Given a list of VertInfoEdges, all all the intra-thread edges (if they exist
-- addAllIntraEdges :: [VertInfoEdge] -> [VertInfoEdge]
-- First, for each item in the passed list calculate if there are any
-- intrathread edges. Then combine the results with the passed list.
--
-- Since addIntraThreadEdge returns Nothing when no intraThreadEdge is used, we
-- map all the nothings to [] and then concat the results
--addAllIntraEdges vs = 
  --(vs ++) . concat . map maybeToList . map (flip addIntraThreadEdge vs) $ vs

-- Given a single VertInfoEdge and a list of VertInfoEdge, attempt to find if
-- an intra-thread predecessor can be added. Given VertInfoEdge 'a', an
-- intra-thread VertInfoEdge can be added if:
--
-- 1. There exists a VertInfoEdge b map such that its snd is from the same
-- thread as fst a
-- 2. snd b is the largest linenumber of all VertInfoEdges satisfying condition
-- one.
--
-- This function returns Nothing if an intra-thread edge cannot be added
--addIntraThreadEdge :: VertInfoEdge -> [VertInfoEdge] -> Maybe VertInfoEdge
--addIntraThreadEdge a bs = 
--  case filterLargestLineNumSnd 
--        . filterHBSnd aFstLN 
--        . filterSameTIdSnd aFstTId 
--        $ bs of
--          [] -> Nothing
--          -- v is a predecessor of a. Return an edge from the second of v to
--          -- the first of a: this is the intra-thread edge. They should both
--          -- have  the same TId already
--          --(v:_) -> Just ((aFstTId, (getVertInfoLineNum . snd $ v)), aFst)
--          (v:_) -> Just ((aFstTId, vSndLN, vSndRW), aFst)
--            where vSnd = snd v
--                  vSndLN = getVertInfoLineNum vSnd
--                  vSndRW = getVertInfoRW vSnd
--  where aFst = fst a
--        aFstTId = getVertInfoTId aFst
--        aFstLN = getVertInfoLineNum aFst

-- Returns a list of VertInfoEdges who's snd vertex is from the same thread as
-- the passed Tid.
--
-- If no such VertInfoEdge is found, returns []
filterSameTIdSnd :: TId -> [VertInfoEdge] -> [VertInfoEdge]
filterSameTIdSnd t vs = filter (vertInfoEdgeSameTIdSnd t) vs

-- Returns true if the snd of the passed VertInfoEdge has the same TId as
-- passed
vertInfoEdgeSameTIdSnd :: TId -> VertInfoEdge -> Bool
vertInfoEdgeSameTIdSnd tid v = vSndTId == tid
  where vSnd = snd v
        vSndTId = getVertInfoTId vSnd

-- Returns the passed [VertInfoEdge] only with those VertInfoEdges who's snd
-- value has a line number less than the lassed LineNum
filterHBSnd :: LineNum -> [VertInfoEdge] -> [VertInfoEdge]
filterHBSnd l vs = filter (vertInfoEdgeHBSnd l) vs

-- Returns true if the snd of the passed VertInfoEdge has a line number less
-- than passed
vertInfoEdgeHBSnd :: LineNum -> VertInfoEdge -> Bool
vertInfoEdgeHBSnd l v = vSndLN < l
  where vSnd = snd v
        vSndLN = getVertInfoLineNum vSnd

-- Returns the passed [VertInfoEdge] only containing the largest snd line
-- number
filterLargestLineNumSnd :: [VertInfoEdge] -> [VertInfoEdge]
filterLargestLineNumSnd [] = []
filterLargestLineNumSnd (v:vs) = filterLargestLineNumSnd' v vs

filterLargestLineNumSnd' :: VertInfoEdge -> [VertInfoEdge] -> [VertInfoEdge]
filterLargestLineNumSnd' v [] = [v]
filterLargestLineNumSnd' oldMax (v:vs) = 
  case (getVertInfoLineNum . snd $ v) > (getVertInfoLineNum . snd $ oldMax) of
    True -> filterLargestLineNumSnd' v vs
    False -> filterLargestLineNumSnd' oldMax vs

-- Given a list of formulas and a map from the variable IDs to their
-- definition, create lists of VertInfoEdges for each formula
createFormulas :: [Formula] -> IdToInfo -> [[VertInfoEdge]]
createFormulas fs m = map (flip createFormula m) fs

createFormula :: Formula -> IdToInfo -> [VertInfoEdge]
createFormula f m = map (varIdToVertInfoEdge m) f

-- Given a variable Id (potentially negated) and a map to look up VertInfoEdges
-- from a VarID, return a VertInfoEdge.
--
-- This will crash if the map does not contain the variable
varIdToVertInfoEdge :: IdToInfo -> VarId -> VertInfoEdge
varIdToVertInfoEdge m v = 
  case v < 0 of
    --True -> flipEdge . (Map.!) m $ (v * (-1))
    --False -> (Map.!) m v
    -- Note: The input formula are the contrainsts which must be added: we
    -- negate them here to figure out what the error causing trace is.
    -- Note: this seems redundant.
    True -> edge
      where edge = m Map.! (v * (-1))
    False -> edge
      where edge = flipEdge . (Map.!) m $ v

-- Given a List of Lists of VertInfoEdges (i.e., a list of Formulas) construct
-- a graph for each formula
--
-- Note: this is lossy since the internal representation of a graph node is
-- only an Int
--
-- The passed IdToInfo should contain the VarId of each VertInfoEdge
createGraphs :: NodeToId -> [[VertInfoEdge]] -> [Graph.Graph]
createGraphs m vs = map (createGraph m) vs

createGraph :: NodeToId -> [VertInfoEdge] -> Graph.Graph
createGraph m vs = Graph.buildG bound (map (getEdge m) vs)
  --where numNodes = getNumNodes vs
    where bound = (1, Map.size m)

-- Given a list of VertInfoEdges return the number of unique nodes across all
-- edges
getNumNodes :: [VertInfoEdge] -> Int
getNumNodes vs = Set.size . Set.fromList . unpair $ vs 

--getNumNodes' :: [VertInfoEdge] -> Set.Set VertInfo -> Int
--getNumNodes' [] acc = undefined
--getNumNodes' (v:vs) acc = undefined

-- Given a VertInfo edge turn it into an edge using the ID found in the passed
-- map
getEdge :: NodeToId -> VertInfoEdge -> Graph.Edge
getEdge m v = (e1, e2)
  where e1 = m Map.! (fst v)
        e2 = m Map.! (snd v)

-- Given a list return a list without duplicates
mkUniq :: Ord a => [a] -> [a]
mkUniq = Set.toList . Set.fromList

-- Given a list of VertInfoEdges return the list without duplicates.
-- This clears the RW flag such that all edges are writes and there will be no
-- duplicates even if in the original formula there were edges which were the
-- same except for their RW flags.
mkEdgeUniqClearRW :: [VertInfoEdge] -> [VertInfoEdge]
mkEdgeUniqClearRW vs = (map makeWrite) . mkUniq $ map dropRW vs
  where dropRW :: VertInfoEdge -> ((TId, LineNum), (TId, LineNum))
        dropRW ((t1,l1,_), (t2,l2,_)) = ((t1,l1), (t2,l2))
        makeWrite :: ((TId, LineNum), (TId, LineNum)) -> VertInfoEdge
        makeWrite ((t1,l1), (t2,l2)) = ((t1,l1, True), (t2,l2, True))

-- Given a list of pairs "undo" all the pairs and return a list of the results
unpair :: [(a, a)] -> [a]
unpair [] = []
unpair ((x1, x2):xs) = x1 : x2 : unpair xs

-- Given a list of edges return a list of all the nodes being used. The will
-- not have any duplicates
getNodes :: [VertInfoEdge] -> [VertInfo]
getNodes vs = mkUniq . unpair $ vs

-- Given a graph, return a list of all the edges which when added to the graph
-- produce a cycle
-- Note: this does not return any self edges
getGraphCyclicEdges :: Graph.Graph -> [Graph.Edge]
getGraphCyclicEdges g = concat . map (getNodeCyclicEdges g) $ (Graph.vertices g)

-- Given a node in the passed graph, return a list of all the edges starting
-- from the passed node which create cycles.
-- Note: this does not create any self edges
getNodeCyclicEdges :: Graph.Graph -> Graph.Vertex -> [Graph.Edge]
getNodeCyclicEdges g n = mkUniq . filterHasCycle g $ edges
  where nodes = Graph.vertices g
        otherNodes = filter ((/=) n) nodes
        edges = getEdges n otherNodes

-- Remove any edges which do not add a cycle to the passed graph
filterHasCycle :: Graph.Graph -> [Graph.Edge] -> [Graph.Edge]
filterHasCycle g es = filterHasCycle' g es []

filterHasCycle' :: Graph.Graph -> [Graph.Edge] -> [Graph.Edge] -> [Graph.Edge]
filterHasCycle' _ [] acc = acc
filterHasCycle' g (e:es) acc = 
  case cyclicGraph newG of
    True -> filterHasCycle' g es (e : acc)
    False -> filterHasCycle' g es acc
  where newG = insertEdge g e 

-- Remove any edges which have a cycle in the passed graph
filterNoCycle :: Graph.Graph -> [Graph.Edge] -> [Graph.Edge]
filterNoCycle g es = filterNoCycle' g es []

filterNoCycle' :: Graph.Graph -> [Graph.Edge] -> [Graph.Edge] -> [Graph.Edge]
filterNoCycle' _ [] acc = acc
filterNoCycle' g (e:es) acc = 
  case cyclicGraph newG of
    False -> filterNoCycle' g es (e : acc)
    True -> filterNoCycle' g es acc
  where newG = insertEdge g e 

-- Returns true if the passed graph has a cycle
cyclicGraph :: Graph.Graph -> Bool
-- For each node, see if it is reachable from itself
cyclicGraph g = or . map (selfReachable g) $ nodes
  where nodes = Graph.vertices g

-- Returns true if there is a path from the passed node to itself in the passed
-- graph
selfReachable :: Graph.Graph -> Graph.Vertex -> Bool
-- Check if there is a path to the passed node from any of its successors
--selfReachable g n = or . map (flip (Graph.path g) n) $ suc
selfReachable g n = selfReachable' g n suc 
  where suc = getSucessors g n
        --pathFlip = flip (Graph.path g) n
        --paths = map pathFlip suc

-- Given a vertex and a list of its sucessesors, return true if there
-- is a path from any successor to the passed vertex. Otherwise, return
-- False
selfReachable' :: Graph.Graph -> Graph.Vertex -> [Graph.Vertex] -> Bool
-- We went through all the vertices and did not find any cyclic paths
-- so return False
selfReachable' _ _ [] = False
selfReachable' g start (n:ns) =
  -- A path from one of the successors (n) to the passed node (start)
  if Graph.path g n start
    then True
    else selfReachable' g start ns

-- Return a list of all the successors of the passed node
getSucessors :: Graph.Graph -> Graph.Vertex -> [Graph.Vertex]
getSucessors g n = map snd . filter (isSource n) $ (Graph.edges g)

-- Returns true if the passed node is the source node of the passed edge
isSource :: Graph.Vertex -> Graph.Edge -> Bool
isSource n e = n == (fst e)

-- Add Intra-thread edges to the passed graph such that they do not create a
-- cycle (i.e., they are valid)
addNonCycIntraThread :: IdToNode -> Graph.Graph -> Graph.Graph
addNonCycIntraThread m g = 
  insertEdges g . filterNoCycle g . getIntraThreadEdges m . Graph.vertices $ g

-- Add Intra-thread edges to the passed graph. This ensures that the returned
-- graph has no cycles (Note: this is expensive)
--
-- The passed map maps a graph node (an Int) a VertInfo (source information).
--addIntraThread :: IdToNode -> Graph.Graph -> Graph.Graph
--addIntraThread m g = 
--  insertEdges g . filterNoCycle g . getIntraThreadEdges m . Graph.vertices $ g

-- Same as addIntraThread except does not filter out cyclic graphs.
addIntraThreadIgnCycles :: IdToNode -> Graph.Graph -> Graph.Graph
addIntraThreadIgnCycles m g = 
  insertEdges g . getIntraThreadEdges m . Graph.vertices $ g

-- Return a list of edges which are edges from two nodes from the same thread
getIntraThreadEdges :: IdToNode -> [Graph.Vertex] -> [Graph.Edge]
getIntraThreadEdges m ns = 
  --filter (lineNumIncreasing m) . filter (intraThread m) . filter (not . selfEdge) . getAllEdges $ ns
  -- TODO: OPTIMIZATION: Only iterate through all the edges once
  getAllIntraThreadIncEdges m ns



-- Returns true if the edge a-->b has an increasing line number (e.g., a is on
-- line 10 and b is in on line 15)
lineNumIncreasing :: IdToNode -> Graph.Edge -> Bool
lineNumIncreasing m e = ln1 < ln2
  where n1 = m Map.! (fst e)
        n2 = m Map.! (snd e)
        --ln1 = snd n1
        --ln2 = snd n2
        ln1 = getVertInfoLineNum n1
        ln2 = getVertInfoLineNum n2

-- Returns true if the passed edge is a self edge
selfEdge :: Graph.Edge -> Bool
selfEdge e = n1 == n2
  where n1 = fst e
        n2 = snd e

-- Returns true if the passed edge is intra-thread
intraThread :: IdToNode -> Graph.Edge -> Bool
intraThread m e = tid1 == tid2
  where n1 = m Map.! (fst e)
        n2 = m Map.! (snd e)
        --tid1 = fst n1
        --tid2 = fst n2
        tid1 = getVertInfoTId n1
        tid2 = getVertInfoTId n2

-- Get all possible edges given the passed list of vertex
getAllEdges :: [Graph.Vertex] -> [Graph.Edge]
--getAllEdges vs = map tuplify2 . replicateM 2 $ vs
getAllEdges vs = [(x,y) | x <- vs, y <- vs]

-- Given the passed list of vertices, generate all possible intra thread edges
-- such that the line numbers within the thread are increasing (e.g., for an
-- edge A->B, the line number of A is greater than the line number of B)
getAllIntraThreadIncEdges :: IdToNode -> [Graph.Vertex] -> [Graph.Edge]
getAllIntraThreadIncEdges m vs = 
  [(x,y) | x <- vs, y <- vs, lineNumIncreasing m (x,y)
                           , intraThread m (x,y)
                           , not . selfEdge $ (x,y)]

-- Dangerous. Convert a list to a 2-tuple assuming the list is a two item list.
-- This will crash if the list does not have two item
tuplify2 :: [a] -> (a,a)
tuplify2 [x,y] = (x,y)
tuplify2 _ = error "List does not contain 2 items"

-- Insert a list of edges to the passed graph
insertEdges :: Graph.Graph -> [Graph.Edge] -> Graph.Graph
insertEdges g [] = g
--insertEdges g (e:es) = insertEdges (insertEdge g e) es
-- TODO: OPTIMIZATION: Instead of inserting the edges one by one, just build a
-- new graph
--insertEdges g (e:es) = insertEdges (insertEdge g e) es
insertEdges g es = Graph.buildG bound newEds 
  where newEds = (Graph.edges g) ++ es 
        bound = (1, length . Graph.vertices $ g)

-- Given a list of edges, create a list of graphs such that each graph is 'g'
-- with one of the passed edges inserted
--getGraphs :: [Graph.Edge] -> Graph.Graph -> [Graph.Graph]
--getGraphs es g = undefined

-- Given a graph, return the graph with the added edge
insertEdge :: Graph.Graph -> Graph.Edge -> Graph.Graph
insertEdge g e = Graph.buildG bound (e : oldEds)
  where oldEds = Graph.edges g 
        bound = (1, length . Graph.vertices $ g)

-- Given a node and a list of nodes create all the edges from the node to each
-- item in the list
getEdges :: Graph.Vertex -> [Graph.Vertex] -> [Graph.Edge]
getEdges n ns = map ((,) n) ns

-- Returns true if the passed edge is between two nodes from the same thread
sameThread :: IdToNode -> Graph.Edge -> Bool
sameThread m e = tid1 == tid2
  where n1 = m Map.! (fst e)
        n2 = m Map.! (snd e)
        --tid1 = fst n1
        --tid2 = fst n2
        tid1 = getVertInfoTId n1
        tid2 = getVertInfoTId n2

-- Convert the nodes from the Graph from (Ints) back into VertInfoEdges using
-- the passed map
edgesToVertInfoEdge :: IdToNode -> [Graph.Edge] -> [VertInfoEdge]
edgesToVertInfoEdge m es = map (edgeToVertInfoEdge m) es

-- Convert a graph to a list of VertInfoEdges using the passed map
graphToVertInfoEdges :: IdToNode -> Graph.Graph -> [VertInfoEdge]
graphToVertInfoEdges m g = edgesToVertInfoEdge m es
  where es = Graph.edges g

-- Given a list of list of edges, return those whos inverse is not in any of
-- the lists. 
-- Note: This may return an empty list when infact the formula is satisfiable
getSatisfiableSimple :: [[Graph.Edge]] -> [Graph.Edge]
-- Check if the swap of an edge is contained in any of the lists of edges
getSatisfiableSimple es = filter (not . swapMember) allEdges
  where allEdges = concat es
        edgeSet = Set.fromList allEdges
        swapMember = (flip Set.member edgeSet) . swap

-- Convert the list of VertInfoEdges to the output format for HBFixer
-- This is of the form:
--
-- <exp1> $ <exp2> $ ... $ <expn>
--
-- Where exp is:
--
-- <line1> <thread ID1>, ..., <lineN> <thread IDN>
toOutput :: [VertInfoEdge] -> String
toOutput vs = concat . intersperse " $ " . map edgeToOutput $ vs

tidToInt :: TId -> Int
tidToInt (TId n) = n

lineToInt :: LineNum -> Int
lineToInt (LineNum n) = n

-- Format VertInfoEdge into format for HBFixer
edgeToOutput :: VertInfoEdge -> String
edgeToOutput v = str
  where n1 = fst v
        n2 = snd v
        --tid1 = tidToInt . fst $ n1
        --tid2 = tidToInt . fst $ n2
        tid1 = tidToInt . getVertInfoTId $ n1
        tid2 = tidToInt . getVertInfoTId $ n2
        --line1 = lineToInt . snd $ n1
        --line2 = lineToInt . snd $ n2
        line1 = lineToInt . getVertInfoLineNum $ n1
        line2 = lineToInt . getVertInfoLineNum $ n2
        str = (show line1) ++ " t" ++ (show tid1) ++ ", " ++ (show line2) 
              ++ " t" ++ (show tid2)

getSatisfiable :: IdToNode -> [[Graph.Edge]] -> [[VertInfoEdge]]
getSatisfiable m es =
  case getSatisfiableSimple es of
    [] -> getSolutions m es 
    es' -> [vis]
      where vis = rmImpliedEdges . onlyInterThread . mkEdgeUniqClearRW . edgesToVertInfoEdge m . mkUniq $ es'

-- Given a list of list of edges, return a "satisfiable" assignment such that
-- at least one item from each sublist is selected.
--
-- The passed map should be a map from a Graph Node ID (Int) to a VertInfoEdge.
--
-- Throws an error if the solution is UNSAT
--getSatisfiable :: IdToNode -> [[Graph.Edge]] -> [[VertInfoEdge]]
--getSatisfiable m es = map mkEdgeUniqClearRW (exploreSolutions m es)
  -- First, try the satisfiablity check before using the sat solver
  --case getSatisfiableSimple es of
    -- Unable to find a simple assignment; use sat solver
    -- Note: The SatSolver returns a list of variables which _all_ must be true
    -- in order for the formula to be satisifed (i.e., they are AND'd
    -- together). 
    --
    -- GetSatifiableSimple returns a list of edges which are OR'd together.
    --
    -- Terms in a product are in the sublists. That is, [[a], [b]] is `a AND
    -- b`. So, convert the sat solver output into a list of lists.
    --[] -> map (:[]) vertSoln
    --  where (form, satVarToVI) = graphEdgesToBoolean m es
    --        soln = solveFormulaUnsafe form (Map.keys satVarToVI)
    --        vertSoln = map ((Map.!) satVarToVI) soln
    --[] -> exploreSolutions m es 
    -- Found simple assignment. This returns a singleton list
    --es' -> [vis]
      --where vis = rmImpliedEdges . onlyInterThread . mkEdgeUniqClearRW . edgesToVertInfoEdge m . mkUniq $ es'

-- Given a set of graph edges to satisfy, find a satisfying assignment to each
-- that does not introduce a cycle
--
-- The input list of lists are those which block the different buggy
-- interleavings. That is, each sublist contains all the edges which introduce
-- a cycle into the interleaving. In order to block the bug, at least one edge
-- must be selected from each sublist.
--
-- This returns a list of lists. Each sublist contains a conjunction (AND) of
-- edges which must be satisfied. The entire list is a list of all the
-- potential solutions found (i.e., selecting any one of the sublists should
-- remove the bug)
getSolutions :: IdToNode -> [[Graph.Edge]] -> [[VertInfoEdge]]
getSolutions m es = rmPermsAndUniq solns 
--getSolutions m es = mkUniq solns 
  where bound = (1, Map.size m)
        -- All possible choices of potential solutions
        --cartProd = sequence es :: [[Graph.Edge]]
        cartProd = nCartProd es :: [[Graph.Edge]]
        -- Graph form of all possible solutions
        gs = map (Graph.buildG bound) cartProd :: [Graph.Graph]
        -- Insert intra-thread edges to the graphs
        fullGs = map (addNonCycIntraThread m) gs :: [Graph.Graph]
        -- Graphs without cycles
        nonCyc = findNUniq (not . cyclicGraph) 100 fullGs
        -- Function to remove non-inter thread, implied, and duplicate edges
        -- from each solution.
        convertAndPrune = onlyInterThread . mkEdgeUniqClearRW 
                          . (graphToVertInfoEdges m)
        -- Convert to edges. 
        --solns = map (graphToVertInfoEdges m) nonCyc
        --soln = rmImpliedEdges . onlyInterThread . mkEdgeUniqClearRW $ graphToVertInfoEdges m nonCyc
        -- This is a list of a bunch of possible solutions
        --solns = map convertAndPrune nonCyc :: [[VertInfoEdge]]
        solns = map convertAndPrune nonCyc :: [[VertInfoEdge]]
        -- Select the "best" solution from the list (we only return one)
        --soln = getBestSolution solns
        --soln = getBestSolution2 solns

-- Get the n-ary cartesian product from the passed list.
-- This is slower but more memory efficient than using sequence.
-- See:
-- https://stackoverflow.com/questions/10122880/lazy-cartesian-product-in-haskell
nCartProd :: [[a]] -> [[a]]
nCartProd  [] = [[]]
nCartProd (x:xs) = [h:t | t <- nCartProd xs, h <- x]

-- Given a list of list of vertinfoedges, where each sublist is an AND of edges
-- which must be satisfied, return the "best" solution.
--
-- We simply favor those solutions which could be solved by a mutex 
getBestSolution :: [[VertInfoEdge]] -> [VertInfoEdge]
getBestSolution [] = []
getBestSolution (a:as) 
    -- If we are at the last item in the list, then we have to return something
    | as == [] = a
    | otherwise = case a of 
                  es@(e1:e2:[]) -> if isMutexSat e1 e2
                                      then es
                                      else getBestSolution as
                  _             -> getBestSolution as

-- The parameter is a list of lists. Each list is a conjunction (AND) of edges
-- which must be satisfied (i.e., a possible bug fix).
--
-- This function searches the possible solutions for the ``best'' fix and
-- returns it. The returned list is in "CNF" form. That is, the sublist(s) are
-- caluses while and each of the clauses are AND'd together.
--
-- The best solution is determined simply based on the number of happens before
-- constraints required. A solution which only requires one edge is favored
-- over a solution which requires two edges.
--
-- In the case where there are solutions requiring one edge, all the solutions
-- are returned and OR'd together. This allows for some optimizations to be
-- used in the repair (e.g., using a mutex).
--
-- When the solution is larger than one, then only one of the solutions is
-- returned. This is due to limitations in the repair implementation
-- (EdgeFixer): it only accepts CNF.
getBestSolution2 :: [[VertInfoEdge]] -> [[VertInfoEdge]]
getBestSolution2 [] = []
getBestSolution2 sols = 
  case (length . head $ sortedSols) of
    -- Gather all the length one solutions and return them in a single list.
    -- This means they are all in one clause
    1 -> [sizeOnes]
      where sizeOnes = concat . (takeWhile isSizeOne) $ sortedSols

    -- Otherwise, just return the first solution. We need to convert it to an
    -- CNF AND: it is a list of unit clauses
    _ -> map (: []) (head sortedSols)
  where sortedSols = sortBy sortListLength sols
        isSizeOne = (==) 1 . length
        sortListLength :: [a] -> [a] -> Ordering
        sortListLength a b = if (length a) < (length b)
                               then LT
                               else if (length a) > (length b)
                                 then GT
                                 else EQ

-- Return true if the two edges could be solved with a mutex

isMutexSat :: VertInfoEdge -> VertInfoEdge -> Bool
isMutexSat e1 e2 = (e1SrcThread == e2SinkThread)
                   && (e1SinkThread == e2SrcThread)
                   -- && (fourOrTwoNodes e1 e2)
                   && (fourNodes e1 e2)
                   -- Check for a cycle
                   && ((e1SrcLine == t1MaxLine)
                      && (e1SinkLine == t2MinLine))
                   && ((e2SrcLine == t2MaxLine)
                      && (e2SinkLine == t1MinLine))
                      where e1Src = fst e1
                            e2Src = fst e2
                            e1Sink = snd e1
                            e2Sink = snd e2
                            e1SrcThread = (getVertInfoTId e1Src)
                            e1SinkThread = (getVertInfoTId e1Sink)
                            e2SrcThread = (getVertInfoTId e2Src)
                            e2SinkThread = (getVertInfoTId e2Sink)
                            e1SrcLine = (getVertInfoLineNum e1Src)
                            e1SinkLine = (getVertInfoLineNum e1Sink)
                            e2SrcLine = (getVertInfoLineNum e2Src)
                            e2SinkLine = (getVertInfoLineNum e2Sink)
                            -- These are split into thread1 and thread2
                            -- although they are not required to be the
                            -- linenumbers associated with each thread.
                            -- However, the first checks above ensure that when
                            -- these values are used they correspond to
                            -- matching threads
                            t1Line1 = getVertInfoLineNum $ e1Src
                            t1Line2 = getVertInfoLineNum $ e2Sink
                            t2Line1 = getVertInfoLineNum $ e1Sink
                            t2Line2 = getVertInfoLineNum $ e2Src
                            t1MinLine = minimum [t1Line1, t1Line2]
                            t1MaxLine = maximum [t1Line1, t1Line2]
                            t2MinLine = minimum [t2Line1, t2Line2]
                            t2MaxLine = maximum [t2Line1, t2Line2]
                            -- Returns true if two edges contain 4 unique nodes
                            -- (i.e., 4 unique line numbers)
                            fourNodes :: VertInfoEdge -> VertInfoEdge -> Bool
                            fourNodes e e' = (Set.size . Set.fromList $ lns) == 4
                              where e1ls = getLineNums e
                                    e2ls = getLineNums e'
                                    lns = e1ls ++ e2ls
                                    getLineNums :: VertInfoEdge -> [LineNum]
                                    getLineNums (vi, vi') = l1 : [l2]
                                      where l1 = getVertInfoLineNum vi
                                            l2 = getVertInfoLineNum vi'

-- same as isMutexSat but accepts a pair of edges. Also, returns Just the
-- passed edges if they are mutex sat. Otherwise, returns nothing
isMutexSatPair :: (VertInfoEdge, VertInfoEdge) -> Maybe (VertInfoEdge, VertInfoEdge)
isMutexSatPair vs@(v1, v2) = 
  if isMutexSat v1 v2
    then Just vs
    else Nothing

-- Unique the list of lists. Permutations of the same list are considered as
-- identical (e.g., [1,2] and [2,1] are the same).
--
-- This is rather expensive since each of the sublists are sorted and then
-- compared to each other. As a result, the order of the sublists is _not_
-- maintained
rmPermsAndUniq :: Ord a => [[a]] -> [[a]]
rmPermsAndUniq l = mkUniq (map sort l)


-- Remove all edges which are implied by other edges in the list
rmImpliedEdges :: [VertInfoEdge] -> [VertInfoEdge]
rmImpliedEdges edges = rmImpliedEdges' edges edges []

-- Carry around the entire original list and an accumulator
rmImpliedEdges' :: [VertInfoEdge] -> [VertInfoEdge] -> [VertInfoEdge] -> [VertInfoEdge]
rmImpliedEdges' _ [] acc = acc
rmImpliedEdges' l (e:es) acc = 
  case isImpliedBy e l of
    True -> rmImpliedEdges' l es acc
    False -> (rmImpliedEdges' l es (e:acc))

-- Returns true if the passed vertInfoEdge is implied by any edge in the list
isImpliedBy :: VertInfoEdge -> [VertInfoEdge] -> Bool
isImpliedBy _ [] = False
isImpliedBy e (v:vs) = 
  if (edgeImplies v e)
    then True
    else isImpliedBy e vs

-- Remove all edges which are implied by the passed edge
rmImpliedBy :: VertInfoEdge -> [VertInfoEdge] -> [VertInfoEdge]
-- Note: we want to keep those edges which are _not_ implied 
rmImpliedBy e es = filter (not . edgeImplies e) es

-- edgeImplies e1 e2: returns true if e1 implies e2. For example, let:
--
-- e1: (5, t1) -> (100, t2)
-- e2: (10, t1) -> (99, t2)
--
-- In this case, e1 implies e2 since whenever e1 is enforced e2 is also
-- enforced. 
edgeImplies :: VertInfoEdge -> VertInfoEdge -> Bool
edgeImplies e1 e2 
  | e1 == e2 = False
  | otherwise = if (happensBefore e1Source e2Source) && (happensBefore e2Sink e1Sink)
                  then True
                  else False
  where e1Source = fst e1
        e2Source = fst e2
        e1Sink = snd e1
        e2Sink = snd e2

-- Returns True if n1 happens before n2 based on the linenumbers of the
-- VertInfos. If there is no relationship between the line numbers (e.g., they
-- are from different threads) returns False.
happensBefore :: VertInfo -> VertInfo -> Bool
happensBefore n1 n2 =
  if (t1 == t2) && (l1 <= l2) 
    then True
    else False
  where t1 = getVertInfoTId n1
        t2 = getVertInfoTId n2
        l1 = getVertInfoLineNum n1
        l2 = getVertInfoLineNum n2


-- Filter any edges which are within the same thread
onlyInterThread :: [VertInfoEdge] -> [VertInfoEdge]
onlyInterThread vs = filter isInterThread vs
  where isInterThread (n1, n2) = (getVertInfoTId n1) /= (getVertInfoTId n2)

-- Same as `find` except throws an error when the item is not found
findUnsafe :: (a -> Bool) -> [a] -> a
findUnsafe f l = case find f l of
                  Just i -> i
                  Nothing -> error "No True item found"

-- Same as find but returns the first `n` items which satisfy the predicate (or
-- less if the end of the list is reached)
findN :: (a -> Bool) -> Int -> [a] -> [a]
findN pr num list = findN' pr num list []
  where findN' :: (a -> Bool) -> Int -> [a] -> [a] -> [a]
        findN' _ _ [] acc = acc
        findN' p n (l:ls) acc 
          | (length acc) == n = acc
          | otherwise = if p l
                          then findN' p n ls (l:acc)
                          else findN' p n ls acc

-- Same as find but returns the first `n` unique items which satisfy the
-- predicate (or less if the end of the list is reached)
findNUniq :: Ord a => (a -> Bool) -> Int -> [a] -> [a]
findNUniq pr num list = findNUniq' pr num list Set.empty
        -- Use a set as the accumulator to keep the found items unique
  where findNUniq' :: Ord a => (a -> Bool) -> Int -> [a] -> Set.Set a -> [a]
        findNUniq' _ _ [] acc = Set.toList acc
        findNUniq' p n (l:ls) acc 
          | (Set.size acc) == n = Set.toList acc
          | otherwise = if p l
                          then findNUniq' p n ls (Set.insert l acc)
                          else findNUniq' p n ls acc

-- Given a list of lists of possible solutions, examine the solutions and add
-- any implied solutions.
--
-- This is done by examining the solutions pairwise and seeing if the
-- combination of the two solutions creates a cycle. This means that the two
-- solutions together represent a critical section (i.e., either A->B or B->A).
-- This means we can add an additional solution which is an OR of two smaller
-- terms.
getImplications :: NodeToId -> IdToNode -> [[VertInfoEdge]] -> [[VertInfoEdge]]
getImplications n2i i2n l = mkUniq (l ++ singles)
  where pairs = mkUniqPairs l
        mutexSolns = catMaybes (map (mutexSolution n2i i2n) pairs)  :: [(VertInfoEdge, VertInfoEdge)]
        -- If mutex solutions are found, they are OR'd together. We want to
        -- append the singleton solutions to the end of all the solutions we've
        -- already found. That is, the mutex solution is [A, B] where either A
        -- or B can be enforced. So, we append [[A], [B]] to represent an OR of
        -- two unit solutions.
        singles = map (: []) (unpair mutexSolns)

-- XXX: UNTESTED
--
-- Convert all the solutions into one big graph and attempt to find a cycle. If
-- a cycle is found, then extract beginging and end of the critical section
--getImplications2 :: NodeToId -> IdToNode -> [[VertInfoEdge]] -> [[VertInfoEdge]]
--getImplications2 n2i i2n l = 
--  if cyclicGraph g
--    then case getCS allSols of
--          Just (v1, v2) -> l ++ [[v1], [v2]]
--          Nothing -> l
--    else l
--  where allSols = concat l
--        g = (addIntraThread i2n) . (createGraph n2i) $ allSols

-- Given a list, return all the pairs of items in the list which are unique
-- (i.e., pairs such that fst /= snd)
mkUniqPairs :: Eq a => [a] -> [(a,a)]
mkUniqPairs l = [(x,y) | x <- l, y <- l, x /= y]

-- Same as mkUniqPairs but only returns the first N pairs
mkNUniqPairs :: Eq a => Int -> [a] -> [(a,a)]
mkNUniqPairs n l = (take n) . mkUniqPairs $ l


-- Given a pair of possible solutions, check if they can be solved via a mutex.
--
-- Two solutions can be solved by a mutex if their combined graph forms a
-- cycle (including the intra-thread edges). Also, the two solutions must only
-- involve two threads (TODO: this can probably be expanded to the three or
-- more thread case)
--
-- If a mutex solution is found, return the two edges making up the critical
-- section. Either one of these edges may be enforced to block the buggy
-- interleaving(s).
--
-- If a mutex solution is not found, returns nothing
--
-- The map is assumed to map each node in the solutions to a unique integer ID
-- (and vice versa)
mutexSolution :: NodeToId -> IdToNode -> ([VertInfoEdge], [VertInfoEdge]) -> Maybe (VertInfoEdge, VertInfoEdge)
mutexSolution n2i i2n (v1, v2) = 
  if cyclicGraph gIntra
    then getCS (v1 ++ v2)
    else Nothing
  where g = createGraph n2i (v1 ++ v2)
        -- TODO: OPTIMIZATION: addIntraThread checks for cycles. So we check
        -- for cycles twice. This is very expensive.
        gIntra = addIntraThreadIgnCycles i2n g

-- Find two edges making up a critical section in the passed list of edges. 
--
-- Edges make up a CS if they start from the maximum line number of one thread
-- and go to the maximum line number of another thread (and vice versa).
--
-- Note: just because these edges exist does not necessairly mean the solution
-- can be solved with a CS. The other condition is that all the edges together
-- make a cycle
--
-- If no such edges are found, then returns Nothing
getCS :: [VertInfoEdge] -> Maybe (VertInfoEdge, VertInfoEdge)
getCS vs = 
  case numThreads vs of
    -- For the two thread case, see if any of the pairs of edges makeup the
    -- mutex pattern
    2 -> case catMaybes (map isMutexSatPair (mkUniqPairs vs)) of
           (v:_) -> Just v
           []     -> Nothing
    _ -> Nothing

-- Given a list of edges, count the number of unique threads
numThreads :: [VertInfoEdge] -> Int
numThreads vs = Set.size . Set.fromList $ ts
  where nodes = unpair vs
        ts = map getVertInfoTId nodes :: [TId]



main :: IO () 
main = do
  -- putStrLn "Starting EdgeMaker"

  -- Open the formulas file
  fFile <- readFile $ formulaFile
  when (length fFile == 0) $ error "No formulas found"

  let formsP = parseFormulaFile . lines $ fFile

  when (not . null . lefts $ formsP) 
      $ error $ "Parse error: " ++ (show (lefts formsP))

  -- Open the variables file
  vFile <- readFile $ variableFile
  when (length vFile == 0) $ error "No formulas found"

  let varsP = parseVariableFile . lines $ vFile

  when (not . null . lefts $ varsP) 
      $ error $ "Parse error: " ++ (show (lefts varsP))

  let forms = rights formsP
  let vars = rights varsP

  -- Map from the ID to VertInfoEdge
  let infoLookup = Map.fromList vars

  -- Given the formulas, expand them into formulas of VertInfoEdges
  --
  -- Add any intraThreadEdges to the formulas as well
  --
  -- It is possible that CBMC returns the ``same'' term in the formula multiple
  -- times. CBMC works at a slightly lower level than the C source code. For
  -- example, a statement `data++;' is a read and write on the same line. Thus,
  -- the output of CBMBC could contain an edge A-->B twice where B is the
  -- modification of data. This can occur if B is a write in one edge and a
  -- read in another. We remove such duplicates with a call to mkUniq.
  --let vertForms = map mkUniq . createFormulas forms $ infoLookup
  let vertForms = map mkEdgeUniqClearRW . createFormulas forms $ infoLookup
  putStrLn $ "Vertex Formulas: " ++ (show vertForms)

  -- Create a map from an ID to each node. This is used because the graph only
  -- supports integer verticies
  let nodes = zip ([1..] :: [Int]) . getNodes . concat $ vertForms
  let idToNode = Map.fromList nodes
  let nodeToId = Map.fromList (map swap nodes)
  putStrLn $ "Node to ID Size: " ++ (show . Map.size $ nodeToId)
  putStrLn $ "NodeID: " ++ (show nodeToId)

  -- Create a graph for each formula
  -- TODO: Do the intrathread edges need to not have cycles?
  --let graphs = map (addNonCycIntraThread idToNode) 
  let graphs = map (addIntraThreadIgnCycles idToNode) 
               . createGraphs nodeToId $ vertForms

  --putStrLn $ "Nodes: " ++ (show nodes)
  --putStrLn $ "Num Nodes: " ++ (show . map (length . Graph.vertices) $ graphs)

  putStrLn $ "Components: " ++ (show .  map (length . Graph.components) $ graphs)
  --putStrLn $ "Graphs: " ++ (show graphs)

  -- Check if there are multiple components per graph
  -- TODO: Can this case legitimately happen?
  -- when ((sum . map (length . Graph.components) $ graphs) /= (length graphs))
  --     $ error "Graph with disconencted components"

  -- Get all the edges causing cycles. Remove any cycles which are coming from
  -- the same thread. Cycles from the same thread are already happening before
  -- each other in the opposite direction.
  let cycEdges =  (map (filter (not . sameThread idToNode) . getGraphCyclicEdges) $ graphs)
  putStrLn $ "Cyclic Edges: " ++ (show cycEdges)
  let ksSize = sum $ map length cycEdges :: Int
  putStrLn $ "Num Cycles: " ++ (show ksSize)

  -- Create a boolean formula of the cyclic edges
  -- The returned map is a map from a SAT variable (Int) to a VertInfoEdge 
  --let (satFormula, satVarToVI) = graphEdgesToBoolean idToNode cycEdges
  --putStrLn $ "SAT Formula: " ++ (show satFormula)
  --putStrLn $ "SAT Map: " ++ (show satVarToVI)

  --let soln = solveFormulaUnsafe satFormula (Map.keys satVarToVI)
  --let vertSoln = map ((Map.!) satVarToVI) soln
  --putStrLn $ "Solution: " ++ (show soln)
  --putStrLn $ "Vert Solution: " ++ (show vertSoln)

  --putStrLn $ "Cyclic Edges: " ++ (show cycEdges)
  --let satEdges = edgesToVertInfoEdge idToNode . mkUniq . getSatisfiableSimple $ cycEdges
  --when ((length satEdges) == 0) (error "Error: No satisfiable edges found")
  ---- putStrLn $ "Satisfiable Cycles: " ++ (show satEdges)
  ---- The output!
  --putStrLn $ (id . toOutput $ satEdges)

  --let (satEdges, numSolns) = getSatisfiable idToNode cycEdges 
  --putStrLn $ unlines (map (id. toOutput) satEdges)
  let solns = getSolutions idToNode cycEdges
  let moreSolns = getImplications nodeToId idToNode solns

  when ((length solns) == 0) (error "Error: UNSAT graph")
  putStrLn $ "Num Solutions: " ++ (show . length $ moreSolns)
  putStrLn $ unlines (map (id. toOutput) moreSolns)

  return ()
