-- Helper functions to convert a set of Graph edges into a boolean formula and
-- then find their satisfying assignment
module GraphEdgeSat (GraphID, 
                     IdToNode, 
                     NodeToId, 
                     graphEdgesToBoolean, 
                     solveFormula,
                     solveFormulaUnsafe,
                     edgeToVertInfoEdge) where 

import qualified Data.Graph as Graph
import Data.Tuple (swap)
import qualified Data.Map.Strict as Map
import Control.Monad.State
--import qualified Data.Boolean as B
import Data.Boolean.SatSolver as S
import Data.Maybe

import VariableParser (VertInfoEdge, VertInfo)

-- ID of a graph node
type GraphID = Int
-- Map from a GraphID to VertInfo and reverse
type IdToNode = Map.Map GraphID VertInfo
type NodeToId = Map.Map VertInfo GraphID


-- A map used to turn GraphEdges into a formula. Provides a lookup from an edge
-- to its integer representation.
--
-- Edges are represented using polarity on an integer. For example, if
-- 1-->(A->B) then -1-->(B->A)
type GraphFormMap = Map.Map Graph.Edge Int

-- State passed around when converting a graph to a formula. Carries a lookup
-- table from a Graph.Edge to its Integer form.
-- Additionally carries around a Map which in the end will map from a SAT
-- variable to its VertInfoEdge.
-- The IdToNode map, which is a map from a Graph ID (Int) to a VertInfo is used
-- to build VertInfoEdges from the Graph edges
-- The counter is used to generate fresh SAT variable names
data GraphFormState = GraphFormState { 
                        edgeToSATVar :: GraphFormMap
                      , graphToVertInfo :: IdToNode
                      , satToVertInfo :: SATVarToVertInfo 
                      , counter :: Int}

type GraphFormStateM = State GraphFormState

type SATVarToVertInfo = Map.Map Int VertInfoEdge

-- Get all the possible solutions of the passed Boolean.
-- The list of Ints is the set of all variables used in the formula.
--
-- Returns a list of lists of ints. Each sublist represents the constraints in
-- one solution (i.e., if an Int appears in a sublist then it is asserted to be
-- true in the solution).
--
-- If the list is empty, this means the formula is UNSAT.
--
-- XXXX: This function may be buggy
getAllSolutions :: Boolean -> [Int] -> [[Int]]
getAllSolutions b vs = getAllSolutions' b vs []

-- Helper function. Carries arround an accumulator of solutions seen so far
--
-- XXXX: This function may be buggy
getAllSolutions' :: Boolean -> [Int] -> [[Int]] ->  [[Int]]
getAllSolutions' b vs acc = 
  case solveFormula b vs of
    -- Conjunct the negation of the solution found to the formula and attempt
    -- to solve it again. Conjuncting with the negation blocks the solution and
    -- forces the solver to try and find another
    Just cs -> getAllSolutions' newB vs newAcc
      where vars = (map ((-1) * ) vars)
            newB = b :&&: (createDisjunct vars)
            newAcc = (cs:acc)
    -- Formula is UNSAT. Return the solutions seen so far
    Nothing -> acc


-- Solve the passed boolean. Returns a list of the constraints which are used
-- in the solution. 
-- The list of Ints provided is the set of all variables used in the formula.
--
-- Returns Nothing if the formula is UNSAT
solveFormula :: Boolean -> [Int] -> Maybe [Int]
solveFormula b vs = do
  -- Create a new SatSolver instance with the passed formula
  satSol <- assertTrue b newSatSolver
  -- Solve the formula
  sol <- solve satSol
  -- Extract the constraints
  return (getConstraints sol vs)

-- Same as solveFormula but throws an error if the formula is UNSAT
solveFormulaUnsafe :: Boolean -> [Int] -> [Int]
solveFormulaUnsafe b vs = case solveFormula b vs of
                            Just is -> is
                            Nothing -> error "Error: UNSAT formula"


-- Given a SatSolver and a list of Ints (variables) return a list containing
-- those Ints which are in the constraints of the solution.
-- Note: the SatSolver should be solved when passed to this function
--
-- If the variable is unconstrained in the solution (i.e., could be either True
-- of False) it is not in the output list
getConstraints :: SatSolver -> [Int] -> [Int]
getConstraints satsol vs = catMaybes $ map (lookupOrNothing satsol) vs
  where lookupOrNothing s i = 
            case S.lookupVar i s of
              Nothing -> Nothing
              Just b -> if b then Just i
                             else Just (i * (-1))


-- Given a map from a GraphNode to a VertInfo, convert a graph edge into a
-- VertInfoEdge
edgeToVertInfoEdge :: IdToNode -> Graph.Edge -> VertInfoEdge
edgeToVertInfoEdge m e = (n1, n2)
  where n1 = m Map.! (fst e)
        n2 = m Map.! (snd e)

-- Convert a list of graph edges into Boolean.
-- The formula is represented as a list of list. The inner list is each caluse
-- (i.e., a sum) and each caluse is conjunected together, e.g.:
--
-- [] AND [] AND [] AND
--
-- Where each list [s1, s2, ...] is :
-- s1 OR s2 OR ...
--
-- Polarity is preserved across edges. So, if the integer 7 represents the edge
-- (1->2), the edge (2->1) is represented by -7
--
-- The IDToNode map is used to create the returned map from variables used in
-- the SAT formula to VertInfoEdges
graphEdgesToBoolean :: IdToNode -> [[Graph.Edge]] -> (S.Boolean, SATVarToVertInfo)
graphEdgesToBoolean gToVI gs = (form, satToVertInfo st)
        -- Initialize the state with IdToNode map so it can be used during the
        -- computation
        -- The counter must start at 1 since there is no negative 0.
  where initState = GraphFormState Map.empty gToVI Map.empty 1
        (ints, st) = runState (graphEdgesToInts gs) initState
        -- For each sublist, convert it into a disjuction (resulting in a list
        -- of Booleans)
        disjuncts = map createDisjunct ints
        -- Create a conjunction out of the entire list of Booleans
        form = createConjunct disjuncts

-- Create a disjunction from a list of Ints. For example:
--
-- [s1, s2, s3] results in (s1 OR s2 OR s3)
createDisjunct :: [Int] -> S.Boolean
createDisjunct [] = S.Yes
createDisjunct is = foldl1 (\acc x -> acc S.:||: x) vars
  where vars = map Var is

-- Create a conjunction from a list of Booleans
--
-- For example [s1, s2, s3] produces (s1 AND s2 AND s3)
createConjunct :: [S.Boolean] -> S.Boolean
createConjunct [] = S.Yes
createConjunct vars = foldl1 (\acc x -> acc S.:&&: x) vars

-- Convert a product-of-sums set of graph edges (pairs of integers) into a
-- boolean formula.
graphEdgesToInts :: [[Graph.Edge]] -> GraphFormStateM [[Int]]
graphEdgesToInts gs = mapM graphEdgesToInts' gs

-- Given a list of graph edges, convert them into a list of Ints representing
-- their formula form
graphEdgesToInts' :: [Graph.Edge] -> GraphFormStateM [Int]
graphEdgesToInts' = mapM insGraphEdge 

-- Given a graph edge, return the value of the edge.
--
-- The state will be updated to reflect a newly seen edge. When an edge is
-- added to the map, both of its polarities are added (i.e., (A->B) and (B->A)
insGraphEdge :: Graph.Edge -> GraphFormStateM Int
insGraphEdge e = 
  do
    -- Map from a Graph edge to a variable in the SAT formula
    gToSAT <- gets edgeToSATVar
    case Map.lookup e gToSAT of
      Just i -> return i
      Nothing -> do
        -- Get the reverse polarity of the edge
        let e2 = swap e
        -- Get a new integer value to represent these edges
        newInt <- gets counter
        modify (\s -> s {counter = newInt + 1 } )
        -- Get the positive and negative faced keys
        let posKey = newInt
        let negKey = newInt * (-1)
        --when (Map.member posKey gToSAT) $ error ("Map collision on key: " ++ (show posKey))
        --when (Map.member negKey gToSAT) $ error ("Map collision on key: " ++ (show negKey))
        -- Insert the two items
        let m1 = Map.insert e posKey gToSAT
        let m2 = Map.insert e2 negKey m1
        -- Insert the new map into the state
        modify (\s -> s { edgeToSATVar = m2 })
        -- Update the SAT variable to vert info map to reflect the new addition
        -- The positive key is associated with the passed edge
        gToVI <- gets graphToVertInfo
        -- Create VertInfoEdges from the passed edge (e) and its swap (e2)
        let vi = edgeToVertInfoEdge gToVI e
        let vi2 = edgeToVertInfoEdge gToVI e2
        sToVI <- gets satToVertInfo
        -- update the map from the SAT variable to the vert info edge
        let sToVI1 = Map.insert posKey vi sToVI
        let sToVI2 = Map.insert negKey vi2 sToVI1
        modify (\s -> s { satToVertInfo = sToVI2 } )
        return (posKey)
