--
-- Author: Markus Kusano
--
-- Parser for the variables file

module VariableParser (parseVariableFile, TId(TId), LineNum(LineNum), VertInfo, VertInfoEdge, IDEdge, flipEdge) where
import Text.ParserCombinators.Parsec
import Control.Applicative ((<$>), (*>), (<*>))
import Data.Tuple (swap)

-- A thread is identified by an integer
newtype TId = TId Int
    deriving(Show, Eq, Ord, Bounded)

-- A line number is just an integer
newtype LineNum = LineNum Int
    deriving(Show, Eq, Ord, Bounded)

-- A vertex in the program is identified by a thread ID and the source code
-- line number and whether it is a read or write. The last boolean is true if
-- the operation is a write
type VertInfo = (TId, LineNum, Bool)

-- An edge of two VertInfos. This indicates that there is a graph edge from fst
-- to snd
type VertInfoEdge = (VertInfo, VertInfo)

-- An Edge along with its ID
type IDEdge = (Int, VertInfoEdge)

-- Given an edge a->b return b->a
flipEdge :: VertInfoEdge -> VertInfoEdge
flipEdge = swap

-- Take a list of each line in a file and parse it into a list of f
parseVariableFile :: [String] -> [Either ParseError IDEdge]
parseVariableFile s = map (parse line "(error)") s

-- Each line of the file contains a bit of data that we want and a lot more
-- data that we just throw away
line :: GenParser Char st IDEdge
line = 
  do
    spaces
    -- First item in the line is the variable ID
    varID <- integer

    -- Parse up to the thread ID
    _ <- string " (e1= [tid"
    (tid1, lineNum1, rw1) <- parseTIdLineNum

    -- Parse to the begining of the next item
    _ <- string ") < (e2= [tid"
    (tid2, lineNum2, rw2) <- parseTIdLineNum

    return (varID, ((TId tid1, LineNum lineNum1, rw1), (TId tid2, LineNum lineNum2, rw2)))

-- Extract the TID and line number, and read--write flag from the happens
-- before edges. This assumes that the parser is up to the point after "tid"
-- (e.g., right before the TId).
--
-- The boolean is true if the operations is a write. Otherwise, false
parseTIdLineNum :: GenParser Char st (Int, Int, Bool)
parseTIdLineNum = 
  do
    threadID <- integer
    _ <- char ']'

    -- Parse (and ignore) the file name
    spaces
    _ <- fileName
    spaces

    -- Parse the line number
    lineNum <- lineNumber
    spaces

    -- Parse the function name and ignore it
    _ <- functionName
    spaces

    -- Parse the read/write flag
    --readWrite <- parseReadWrite
    --spaces

    ---- Parse the addr and ignore it
    --_ <- parseAddr
    --spaces
    --return (threadID, lineNum, readWrite)

    -- Parse the read/write flag and ignore the addr
    (readWrite, _) <- parseReadWriteAddr 
    spaces
    return (threadID, lineNum, readWrite)

-- The filename starts with 'file' followed by a space and then the filename
-- TODO: if the filename has a space in it this might have problems
fileName :: GenParser Char st String
fileName = 
  do
    _ <- string "file "
    many (noneOf " ")

-- The line number starts with "line" and then an integer surrounded by spaces
lineNumber :: GenParser Char st Int
lineNumber = 
  do
    _ <- string "line "
    integer

-- The function name started with "function" followed by the name followed by a
-- space
functionName :: GenParser Char st String
functionName = 
  do
    _ <- string "function "
    many (noneOf " ")

-- Parse the string "read" or "write" to get the read/write flag of an event in
-- a happens-before edge
parseReadWrite :: GenParser Char st Bool
parseReadWrite = 
  do
    rwStr <- string "write" <|> string "read" 
    if (rwStr == "read")
    then return False
    else return True

-- Parse the string "read" or "write" to get the read/write flag of an event in
-- a happens-before edge. Then parse the addr. Note: the addr: may be empty
-- (not sure why CBMC does this).
parseReadWriteAddr :: GenParser Char st (Bool, String)
parseReadWriteAddr =
  do
    rwStr <- string "write" <|> string "read"  <|> string "addr:"
    if (rwStr == "addr:")
      then 
        do 
          spaces
          return (True, "")
      else 
        do 
          spaces
          addr <- parseAddr
          if (rwStr == "read")
            then return (False, addr)
            else return (True, addr)
    

-- Not too sure what this is. It is either "  addr" or " addr"
-- followed by some other stuff
parseAddr :: GenParser Char st String
parseAddr = 
  do
    --_ <- string "write" <|> string "read" 
    --spaces
    _ <- string "addr:"
    spaces
    a1 <- many (noneOf ":")
    _ <- string "::"
    a2 <- many (noneOf " ")
    return (a1 ++ "::" ++ a2)

        


-- Parse an integer
integer :: GenParser Char st Int
integer = rd <$> (plus <|> minus <|> number)
    where rd     = read :: String -> Int
          plus   = char '+' *> number
          minus  = (:) <$> char '-' <*> number
          number = many1 digit
