/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: http://gcc.gnu.org/bugzilla/show_bug.cgi?id=3584
 *
 * Atomicity violation in ARM based atomic operations in GCC's libstdc++
 *
 * See README.md for more information
 *
 * Note: This program was written on an x86 machine and it has not been tested
 */

#include <pthread.h>
#include <stdint.h>
#include <assert.h>

#define INT_TYPE int8_t

/**
 * Global value that is manipulated by threads in t_main()
 */
INT_TYPE val;

/**
 * Calls atomic add on val to add 2 to it
 *
 * \param args unused
 * \return Returns NULL upon completion
 */
void *t1_main(void *args);
void *t2_main(void *args);

// Increment the value stored at the passed memory location by the passed
// ammount
void atomic_add_t1(int *p, int amt);
// Increment the value stored at the passed memory location by the passed
// ammount
void atomic_add_t2(int *p, int amt);

void *t1_main(void *args) {

  int v = 2;

  atomic_add_t1(&val, v);

  return NULL;
}

void atomic_add_t1(int *p, int amt) {
  // These three operations (read-modify-write) should be atomic but they are
  // not
  // In the actual GCC code, these are implemented as inline ARM assembly.
  int tmp = *p;
  tmp += amt;
  *p = tmp;
}

void *t2_main(void *args) {
  int v = 2;

  atomic_add_t2(&val, v);

  return NULL;
}

void atomic_add_t2(int *p, int amt) {
  int tmp = *p;
  tmp += amt;
  *p = tmp;
}

/**
 * Creates two threads in t{1,2}_main() after initializing val to 0.
 * Then, checks if the final state of val is 4.
 *
 * \return Returns 0 upon completion
 */
int main() {
    pthread_t t1, t2;

    // Initialize val
    val = 0;

    // Create threads
    pthread_create(&t1, NULL, t1_main, NULL);
    pthread_create(&t2, NULL, t2_main, NULL);

    // Wait for the threads to finish
    pthread_join(t1, NULL);
    pthread_join(t2, NULL);

    // Since both threads add 2 to val, the correct final state should be 4.
    // However, since they are racing to read-modify-write val, the end result
    // can be 2.
    if (val != 4) {
      assert(0);
    }

    return 0;
}
