/**
 * \file jetty_atom_violation.c
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Example of an atomicity violation found in Jetty. Two threads increment a
 * global variable without synchronization which results in a potential for the
 * non-atomic increments to overlap on the assembly level.
 *
 * See README.md for more information
 */

#include <pthread.h>

// Comment out to remove the call to assert() in main
// #define ASSERT

#ifdef ASSERT
#include <stdio.h>
#include <assert.h>
#endif

/**
 * Global variable that is raced after
 */
int _set;
pthread_mutex_t m = PTHREAD_MUTEX_INITIALIZER;

/**
 * Increments _set
 */
void *t1_main(void *args);

/**
 * Decrements _set but holds a lock while doing so
 */
void *t2_main(void *args);

void *t1_main(void *args) {
  _set++;
  return NULL;
}

void *t2_main(void *args) {
  pthread_mutex_lock(&m);
	_set--;
  pthread_mutex_unlock(&m);
  return NULL;
}

/**
 * Initializes _set to zero and spawns three threads in t_main(). Waits for the
 * threads to return and then asserts that _set == 0 
 *
 *
 * \return Returns 0 upon completion
 */
int main(void) {
    pthread_t t1, t2, t3, t4;

    // Initialize set to zero
    _set = 0;

    // Spawn threads into t_main
    pthread_create(&t1, NULL, t1_main, NULL);
    pthread_create(&t2, NULL, t2_main, NULL);


    pthread_join(t1, NULL);
    pthread_join(t2, NULL);

    assert(_set == 0);

    return 0;
}
