/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 *
 * Source: http://llvm.org/bugs/show_bug.cgi?id=8441
 *
 * Simple example of a data race and atomicity violation  found in LLVM. The
 * structure AttributesList is destroyed when its reference count is zero.
 * There is no synchronization around when the structure's reference count is
 * checked to be zero and the subsequent deletion of the structure.
 *
 * See README.md for more information
 */

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <assert.h>

#include <unistd.h>

#define UINT_TYPE uint8_t

/// Typedef for the AttributeList structure. It simply only contains a reference
// count and some data. 
#if 0
typedef struct AttributeList {
    int ref_count;  ///< The Reference count for the structure. The structure
		    /// can safely be deleted when the reference count is zero
    char *data;	    ///< Data store in the structure
} AttributeList;    

// Global AttributeList that is raced after
AttributeList *att_list;
#endif

// reference counted data structure
UINT_TYPE ref_count_g;
char data_g;
bool destroyed_t1_g;
bool destroyed_t2_g;

/// Decrements the passed AttributeList's reference count. If the reference
// count becomes zero then the structure can be deleted. It is assumed that the
// data memeber of att_list has been allocated with a call to malloc() since it
// will be free()'ed
//
//  \param att_list Pointer to AttributeList to decrement the reference count
//  of
void dec_att_list_ref_count_t1();
void dec_att_list_ref_count_t2();

/// Increments the passed AttributeList's ref_count
//
// \param att_list Pointer to AttributeList to increment the ref_count of
void inc_att_list_ref_count_t1();
void inc_att_list_ref_count_t2();

/// Allocates memory and initializes the passed att_list. The data section is
// given the size of one char of memory
//
// \param att_list Attribute List to initialize
void att_list_init();

/// Free()'s the memory of the data section of the passed att_list and then
// free()'s the AttributeList itself
//
// \param att_list Pointer to AttributeList to destroy
void att_list_destroy_t1();
void att_list_destroy_t2();

/// Increments the global AttributeList att_list's ref_count, manipulates it's
// data section and then decrements it's ref_count
//
// \param args Unused
// \return Returns NULL upon completion
void *t1_main(void *args);
void *t2_main(void *args);

void att_list_init() {
#if 0
    att_list = malloc(sizeof(AttributeList));

    if (att_list == NULL) {
      fprintf(stderr, "Error: Call to malloc() returned NULL\n");
      exit(EXIT_FAILURE);
    }

    att_list->data = malloc(sizeof(char));

    att_list->ref_count = 0;
    *att_list->data = 'a';
#endif
    data_g = 'x';
    ref_count_g = 0;
    destroyed_t1_g = false;
}

// Begin thread one functions
void *t1_main(void *args) {
    // Increment the reference count to refelect the strucutre being used
    //inc_att_list_ref_count_t1();
    ref_count_g++;

    // Modify the data section
    data_g = 'a';

    // We're done using the structure so decrement the ref_count
    //dec_att_list_ref_count_t1();
    ref_count_g--;

    if (ref_count_g == 0) {
      att_list_destroy_t1();
    }

    return NULL;
}

void inc_att_list_ref_count_t1() {
#if 0
    #ifdef DEBUG
    printf("Incrementing reference count\n");
    #endif

    att_list->ref_count++;
#endif
    ref_count_g++;
}

void dec_att_list_ref_count_t1() {
#if 0
    if (att_list->ref_count == 0) {
      fprintf(stderr, "Warning: Attempting to decrement the reference count" 
            "of an AttributeList that already has a 0 ref_count\n");
      att_list_destroy(att_list);
      return;
    }

    att_list->ref_count--;

    if (att_list->ref_count == 0) {
      #ifdef DEBUG
      printf("ref_count == 0, free()'ing structure\n");
      #endif
      att_list_destroy(att_list);
      // Oops, no atomicity between the check and the call to destroy
    }
#endif
    ref_count_g--;

    if (ref_count_g == 0) {
      destroyed_t1_g = true;
    }

}

void att_list_destroy_t1() {
#if 0
    free(att_list->data);
    att_list->data = NULL;
    free(att_list);
    att_list = NULL;
#endif
    destroyed_t1_g = true;
}
// End thread one functions

// Begin thread two functions
void *t2_main(void *args) {
    // Increment the reference count to refelect the strucutre being used
    //inc_att_list_ref_count_t2();
    ref_count_g++;

    // Modify the data section
    data_g = 'a';

    // We're done using the structure so decrement the ref_count
    //dec_att_list_ref_count_t2();
    ref_count_g--;

    if (ref_count_g == 0) {
      destroyed_t2_g = true;
    }

    return NULL;
}

void inc_att_list_ref_count_t2() {
#if 0
    #ifdef DEBUG
    printf("Incrementing reference count\n");
    #endif

    att_list->ref_count++;
#endif
    ref_count_g++;
}

void dec_att_list_ref_count_t2() {
#if 0
    if (att_list->ref_count == 0) {
      fprintf(stderr, "Warning: Attempting to decrement the reference count" 
            "of an AttributeList that already has a 0 ref_count\n");
      att_list_destroy(att_list);
      return;
    }

    att_list->ref_count--;

    if (att_list->ref_count == 0) {
      #ifdef DEBUG
      printf("ref_count == 0, free()'ing structure\n");
      #endif
      att_list_destroy(att_list);
      // Oops, no atomicity between the check and the call to destroy
    }
#endif
    ref_count_g--;

    if (ref_count_g == 0) {
      att_list_destroy_t2();
    }

}

void att_list_destroy_t2() {
#if 0
    free(att_list->data);
    att_list->data = NULL;
    free(att_list);
    att_list = NULL;
#endif
    destroyed_t2_g = true;
}
// End thread two functions
/// Initializes the global AttributeList att_list and then creates two threads
// in t_main()
//
// \return Return's 0 upon completion
int main(void) {
    pthread_t t1;
    pthread_t t2;

    att_list_init();

    pthread_create(&t1, NULL, t1_main, NULL);
    pthread_create(&t2, NULL, t2_main, NULL);

    //pthread_join(t1, NULL);
    //pthread_join(t2, NULL);

    if (destroyed_t1_g && destroyed_t2_g) assert(0);

    return 0;
}

