/**
 * \file
 * \author Markus Kusano (mukusano@vt.edu)
 * Program exemplifying a  found in MySQL-4.0.26.
 *
 * Source: Source: http://web.eecs.umich.edu/~jieyu/bugs/mysql-12848.html
 * MySQL Source: 
 *
 * See README.md for details
 */

#include <pthread.h>
#include <stdlib.h>
#include <stdio.h>

// Comment out to disable debugging messages
// #define DEBUG

/** 
 * A simplified representation of the cache in the original program
 */
struct cache {
    char *data;	/**< Data stored in the cache */
    int size;	/**< Number of bytes written to data */
} Cache;    /**< Instance of struct cache for the program to use */

/**
 * Mutex to protect the cache size.
 */
pthread_mutex_t guard_mutex = PTHREAD_MUTEX_INITIALIZER;

/**
 * Free memory (cache.data)
 */
void cache_free();

/**
 * Disables the cache. This is represented by setting the size to 0.
 */
void cache_make_disable();

/**
 * Resize the cache to newSize
 * 
 * \param newSize size the cache should be adjusted to
 */
void cache_resize(int newSize);

/**
 * Initializes the cache. Allocates memory for cache.data that is of the size
 * of cache.size
 */
void cache_init();

/**
 * Asign the passed query to Cache.data
 *
 * \param query Data to store in the cache
 */
void cache_store_query(char *query);

/* Thread Functions */
/**
 * Insert a query into the cache. 
 *
 * \param query Char* to be stored in cache.data
 */
void *querier(void *query);

void cache_free() {
    pthread_mutex_lock(&guard_mutex);
    free(Cache.data);
    Cache.data = NULL;
    cache_make_disable();
    pthread_mutex_unlock(&guard_mutex);
}

void cache_make_disable() {
    Cache.size = 0;
}

void cache_resize(int newSize) {
    // Manually inlined
    // This entire section should be protected to ensure atomicity
    //cache_free();
    free(Cache.data);
    Cache.data = NULL;

    //cache_make_disable();
    Cache.size = 0;


    //Cache.size = newSize;
    // Uncomment the following to cause the program to segfault, showing the
    // buggy interleaving
    // sleep(1);
    //cache_init();
    //Cache.data = malloc(sizeof(char)*Cache.size);
}

void cache_init() {
    Cache.data = malloc(sizeof(char)*Cache.size);
}

void cache_store_query(char *query) {
  if (Cache.size > 0) {
    assert(Cache.data != NULL);
    *(Cache.data) = *query;
  }
}

void *querier(void *unused) {
    //cache_store_query((char *) query);
  if (Cache.size > 0) {
    // Check for atomicity in the other thread
    assert(Cache.data != NULL);
  }
}

/**
 * Sets the cache sie to 64 and initializes it via cache_init. A thread is
 * created in querier*() and then the cache is resized
 *
 * \return Returns 0 upon completion
 */
int main() {
    pthread_t t1, t2;

    // Allocate data to the cache
    Cache.size = 64;
    cache_init();

    // Have one thread resize the cache to size 128
    pthread_create(&t1, NULL, querier, NULL);

    // Have main act as the other thread to resize the cache
    cache_resize(128);

    pthread_exit((void *) 0);
}


